<#
.Synopsis
    Restores databases from Ola Hallengren's Maintenance Solution backups to the target server.
.DESCRIPTION
    This script is for use when a full restore of all databases on a server is required, such as
    in an emergency situation. The script will parse the directory passed into the parameter $bkdr
    and then into its subdirectory containing the target server name. For each non-system database it
    finds, it will test to see if a full backup exists for that database, and if so, will start the
    restore process.
    
    If the database already exists on the target server, it will put that database in single-user mode
    so it can cleanly restore the database. It then restores the database from the most recent full
    backup file it finds, with norecovery. Next it looks for the most recent differential backup for
    the database, and if found, restores that, with norecovery. It then retrieves all of the transaction
    log backups after the most recent restored backup, and restores them in order, each with norecovery.
    
    Once the restores have been completed, the script will 'recover' the database, and set it into multi-
    user mode.
    
.EXAMPLE
   ./Restore-MaintSolution.ps1 -tgtsvr WS12SQL -bkdr 'C:\Work'

#>

#	Revision History---------------------------------------------------
#	v1.00	Allen White		23 October, 2019

# Get the parameters from the command line
[CmdletBinding()] 
param(
  # $tgtsvr is the target SQL Server instance
  [Parameter(Mandatory=$true)]
  [string]$tgtsvr=$null,
  # $bkdr is the directory path containing the backup files
  [Parameter(Mandatory=$true)]
  [string]$bkdr=$null
  )

try {
    $connstr = "server=$($tgtsvr);database=master;Application Name=RestoreDB;Integrated Security=true;"
    $cn = new-object System.Data.SqlClient.SqlConnection($connstr);

    $bckdb = Get-ChildItem "$($bkdr)/$($tgtsvr)" | Where {$_.Name -ne 'master' -and $_.Name -ne 'model' -and $_.Name -ne 'msdb' -and $_.Name -ne 'ReportServerTempDB'}
    foreach ($db in $bckdb) {

        $dbnm = $db.Name
        $fullfile = Get-ChildItem "$($bkdr)/$($tgtsvr)/$($dbnm)/FULL/" | Sort-Object LastWriteTime -Desc | Select-Object -First 1

        # Make sure there's a full backup file for the database before doing anything else
        if ($fullfile.Count -eq 1) {
            $lastbk = $fullfile.LastWriteTime
            $bkupfile = $fullfile.FullName

            # Open the Connection object and create a SqlCommand object
            $cn.Open();
            $cmd = New-Object System.Data.SqlClient.SqlCommand
            $cmd.Connection = $cn

            # Check to see if the database already exists
            $q = "SELECT Name from sys.databases WHERE Name = '$($dbnm)'"
            $cmd.CommandText = $q
            $dr = $cmd.ExecuteReader()
            while ($dr.Read()) { 
                $dbname = $dr.GetValue(0)
                }
            $dr.Close()

            # If so, set it in single user mode
            if ($dbnm -eq $dbname) {
                $q = "ALTER DATABASE [$($dbnm)] SET SINGLE_USER WITH ROLLBACK IMMEDIATE"
                $cmd.CommandText = $q
                $result = $cmd.ExecuteNonQuery();
                }

            $q = "RESTORE DATABASE [$($dbnm)] FROM  DISK = '$($bkupfile)' WITH  FILE = 1,  NOUNLOAD,  REPLACE, NORECOVERY"
            $cmd.CommandText = $q
            $result = $cmd.ExecuteNonQuery();

            if (Test-Path -Path "$($bkdr)/$($tgtsvr)/$($dbnm)/DIFF/") {
                $difffile = Get-ChildItem "$($bkdr)/$($tgtsvr)/$($dbnm)/DIFF/" | Where-Object {$_.LastWriteTime -gt $lastbk} | Sort-Object LastWriteTime -Desc | Select-Object -First 1
                $lastbk = $difffile.LastWriteTime
                $bkupfile = $difffile.FullName

                $q = "RESTORE DATABASE [$($dbnm)] FROM  DISK = '$($bkupfile)' WITH  FILE = 1,  NOUNLOAD,  REPLACE, NORECOVERY"
                $cmd.CommandText = $q
                $result = $cmd.ExecuteNonQuery();
                }
            if (Test-Path -Path "$($bkdr)/$($tgtsvr)/$($dbnm)/LOG/") {
                $tranfile = Get-ChildItem "$($bkdr)/$($tgtsvr)/$($dbnm)/LOG/" | Where-Object {$_.LastWriteTime -gt $lastbk} | Sort-Object LastWriteTime
                foreach ($trnfile in $tranfile) {
                    $bkupfile = $trnfile.FullName

                    $q = "RESTORE LOG [$($dbnm)] FROM  DISK = '$($bkupfile)' WITH  FILE = 1,  NOUNLOAD,  REPLACE, NORECOVERY"
                    $cmd.CommandText = $q
                    $result = $cmd.ExecuteNonQuery();
                    }
                }
            $q = "RESTORE DATABASE [$($dbnm)] WITH RECOVERY"
            $cmd.CommandText = $q
            $result = $cmd.ExecuteNonQuery();

            $q = "ALTER DATABASE [$($dbnm)] SET MULTI_USER"
            $cmd.CommandText = $q
            $result = $cmd.ExecuteNonQuery();

            $cn.Close()

            }
        }
    }

catch {
    # Handle the error
    $err = $_.Exception
    write-output $err.Message
    while( $err.InnerException ) {
        $err = $err.InnerException
        write-output $err.Message
        }
    }
finally {
    write-output "script completed"
    }